#include "fresh.h"

#include <Arduino.h>
#include <time.h>
#include <EEPROM.h>

/**
 * @brief The allowed discrepancy between the time of a received message
 * and the device time (in seconds)
 * 
 * A stricter (lower) value better prevents against replay attacks,
 * but may lead to issues when dealing with slow networks and other
 * routing delays.
 */
uint32_t allowedOffset = 60;

void setMessageTimeAllowedOffset(uint32_t offset) {
    allowedOffset = offset;
}

void configureNTP(int32_t offsetToGMT, int32_t offsetDaylightSavings, const char* serverUrl) {
    configTime(offsetToGMT, offsetDaylightSavings, serverUrl);
}

void printLocalTime() {
    struct tm timeinfo;
    if (getLocalTime(&timeinfo)) {
        Serial.println(&timeinfo, "[INFO] Time is %A, %d. %B %Y %H:%M:%S");
    } else {
        Serial.println("[WARN] No local time available");
    }
}

uint32_t getEpochTime() {
    time_t now;
    struct tm timeinfo;
    if (!getLocalTime(&timeinfo)) {
        Serial.println("[WARN] Failed to obtain local time");
        return(0);
    }
    time(&now);
    return now;
}

bool isMessageTimeAcceptable(uint32_t t) {
    uint32_t localTime = getEpochTime();
    if (localTime == 0) {
        Serial.println("No epoch time available");
        return false;
    }
    if (t > localTime + allowedOffset) {
        return false;
    }
    if (t < localTime - allowedOffset) {
        return false;
    }
    return true;
}

void prepareMessageCounterUsage() {
    EEPROM.begin(MESSAGE_COUNTER_SIZE * remoteDeviceCount);
}

uint32_t getNextMessageCounter(uint8_t deviceId) {
    int offset = deviceId * MESSAGE_COUNTER_SIZE;
    uint32_t counter = (uint32_t) EEPROM.read(offset + 0) << 24;
    counter         += (uint32_t) EEPROM.read(offset + 1) << 16;
    counter         += (uint32_t) EEPROM.read(offset + 2) << 8;
    counter         += (uint32_t) EEPROM.read(offset + 3);
    return counter;
}

void printMessageCounters() {
    Serial.print("[INFO] Next message numbers:");
    for (uint8_t i = 0; i < remoteDeviceCount; i += 1) {
        Serial.printf(" %u", getNextMessageCounter(i));
    }
    Serial.println("");
}

bool isDeviceIdValid(uint8_t deviceId) {
    return deviceId < remoteDeviceCount;
}

bool isMessageCounterValid(uint32_t counter, uint8_t deviceId) {
    return counter >= getNextMessageCounter(deviceId);
}

void setMessageCounter(uint32_t counter, uint8_t deviceId) {
    int offset = deviceId * MESSAGE_COUNTER_SIZE;
    EEPROM.write(offset + 0, (counter >> 24) & 0xFF);
    EEPROM.write(offset + 1, (counter >> 16) & 0xFF);
    EEPROM.write(offset + 2, (counter >>  8) & 0xFF);
    EEPROM.write(offset + 3,  counter        & 0xFF);
    EEPROM.commit();
}

void didUseMessageCounter(uint32_t counter, uint8_t deviceId) {
    // Store the next counter, so that resetting starts at 0
    setMessageCounter(counter+1, deviceId);
}

void resetMessageCounters() {
    for (uint8_t i = 0; i < remoteDeviceCount; i += 1) {
        setMessageCounter(0, i);
    }
    Serial.println("[WARN] Message counters reset");
}
