/**
 * Sesame-Device
 * Christoph Hagen, 2022.
 *
 * The code for a simple door unlock mechanism where a servo pushes on an existing
 * physical button.
 */
#include <Arduino.h>
#include <WiFi.h>
#include <ESPAsyncWebServer.h>

#include "crypto.h"
#include "fresh.h"
#include "message.h"
#include "server.h"
#include "servo.h"
#include "config.h"

/* Global variables */

ServerConnection server(serverUrl, serverPort, serverPath);

ServoController servo(pwmTimer, servoFrequency, servoPin);

AsyncWebServer local(localPort);

// The buffer to hold a received message while it is read
uint8_t receivedMessageBuffer[AUTHENTICATED_MESSAGE_SIZE];

/* Event callbacks */

SesameEvent handleReceivedMessage(AuthenticatedMessage* payload, AuthenticatedMessage* response);

// Forward declare monitoring functions
void ensureWiFiConnection(uint32_t time);
void ensureWebSocketConnection(uint32_t time);

void sendFailureResponse(AsyncWebServerRequest *request, SesameEvent event);
void sendMessageResponse(AsyncWebServerRequest *request, SesameEvent event, AuthenticatedMessage* message);
void sendResponse(AsyncWebServerRequest *request, uint8_t* buffer, uint8_t size);
void hexToBin(const char * str, uint8_t * bytes, size_t blen);

/* Logic */

void setup() {
    Serial.begin(serialBaudRate);
    Serial.setDebugOutput(true);
    Serial.println("[INFO] Device started");

    servo.configure(lockOpeningDuration, servoPressedState, servoReleasedState);
    Serial.println("[INFO] Servo configured");

    prepareMessageCounterUsage();
    //resetMessageCounters();
    printMessageCounters();

    server.onMessage(handleReceivedMessage);

    local.on("/message", HTTP_POST, [] (AsyncWebServerRequest *request) {
        if (!request->hasParam(messageUrlParameter)) {
            Serial.println("Missing url parameter");
            sendFailureResponse(request, SesameEvent::InvalidMessageData);
            return;
        }
        String encoded = request->getParam(messageUrlParameter)->value();
        hexToBin(encoded.c_str(), receivedMessageBuffer, AUTHENTICATED_MESSAGE_SIZE);
        // Process received message
        AuthenticatedMessage* message = (AuthenticatedMessage*) receivedMessageBuffer;
        AuthenticatedMessage responseMessage;
        SesameEvent event = handleReceivedMessage(message, &responseMessage);
        sendMessageResponse(request, event, &responseMessage);
    });
}

void loop() {
    uint32_t time = millis();

    server.loop();
    servo.loop();

    ensureWiFiConnection(time);
    ensureWebSocketConnection(time);
}

uint32_t nextWifiReconnect = 0;
bool isReconnecting = false;

void ensureWiFiConnection(uint32_t time) {
    // Reconnect to WiFi
    if(time > nextWifiReconnect && WiFi.status() != WL_CONNECTED) {
        Serial.println("[INFO] Reconnecting WiFi...");
        WiFi.setHostname(networkName);
        WiFi.begin(wifiSSID, wifiPassword);
        isReconnecting = true;
        nextWifiReconnect = time + wifiReconnectInterval;
    }
}

void ensureWebSocketConnection(uint32_t time) {
    if (isReconnecting && WiFi.status() == WL_CONNECTED) {
        isReconnecting = false;
        Serial.println("IP address: ");
        Serial.println(WiFi.localIP());
        Serial.println("[INFO] WiFi connected, opening socket");
        server.connectSSL(serverAccessKey);
        configureNTP(timeOffsetToGMT, timeOffsetDaylightSavings, ntpServerUrl);
        printLocalTime();
        local.begin();
    }
}

SesameEvent processMessage(AuthenticatedMessage* message) {
    if (!isDeviceIdValid(message->message.device)) {
        return SesameEvent::MessageDeviceInvalid;
    }
    if (!isMessageCounterValid(message->message.id, message->message.device)) {
        return SesameEvent::MessageCounterInvalid;
    }
    if (!isMessageTimeAcceptable(message->message.time)) {
        return SesameEvent::MessageTimeMismatch;
    }
    if (!isAuthenticMessage(message, remoteKey, keySize)) {
        return SesameEvent::MessageAuthenticationFailed;
    }
    return SesameEvent::MessageAccepted;
}

/**
 * Process received binary data.
 *
 * Checks whether the received data is a valid and unused key,
 * and then signals that the motor should move.
 *
 * @param payload The pointer to the received data.
 * @param length The number of bytes received.
 * @return The event to signal to the server.
 */
SesameEvent handleReceivedMessage(AuthenticatedMessage* message, AuthenticatedMessage* response) {
    SesameEvent event = processMessage(message);
    
    // Only open when message is valid
    if (event == SesameEvent::MessageAccepted) {
        didUseMessageCounter(message->message.id, message->message.device);
        // Move servo
        servo.pressButton();
        Serial.printf("[Info] Accepted message %d\n", message->message.id);
    }

    // Create response for all cases
    response->message.time = getEpochTime();
    response->message.id = getNextMessageCounter(message->message.device);
    if (!authenticateMessage(response, localKey, keySize)) {
        return SesameEvent::MessageAuthenticationFailed;
    }
    return event;
}

void sendFailureResponse(AsyncWebServerRequest *request, SesameEvent event) {
    uint8_t response = static_cast<uint8_t>(event);
    sendResponse(request, &response, 1);
}

void sendMessageResponse(AsyncWebServerRequest *request, SesameEvent event, AuthenticatedMessage* message) {
    uint8_t response[AUTHENTICATED_MESSAGE_SIZE+1];
    response[0] = static_cast<uint8_t>(event);
    memcpy(response+1, (uint8_t*) message, AUTHENTICATED_MESSAGE_SIZE);
    sendResponse(request, response, AUTHENTICATED_MESSAGE_SIZE+1);
}

void sendResponse(AsyncWebServerRequest *request, uint8_t* buffer, uint8_t size) {
    request->send_P(200, "application/octet-stream", buffer, size);
    Serial.printf("[INFO] Local response %d\n", buffer[0]);
}

// Based on https://stackoverflow.com/a/23898449/266720
void hexToBin(const char * str, uint8_t * bytes, size_t blen) {
    uint8_t  idx0, idx1;

    // mapping of ASCII characters to hex values
    const uint8_t hashmap[] = {
        0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, // 01234567
        0x08, 0x09, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // 89:;<=>?
        0x00, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f, 0x00, // @ABCDEFG
        0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // HIJKLMNO
    };

    memset(bytes, 0, blen);
    
    size_t len = strlen(str);
    if (len % 2) {
        // Require two chars per byte
        return;
    }
    size_t end = min(blen*2, len);

    for (size_t pos = 0; pos < end; pos += 2) {
        idx0 = ((uint8_t)str[pos+0] & 0x1F) ^ 0x10;
        idx1 = ((uint8_t)str[pos+1] & 0x1F) ^ 0x10;
        bytes[pos/2] = (uint8_t)(hashmap[idx0] << 4) | hashmap[idx1];
    };
}