#include "server.h"

constexpr int32_t pingInterval = 10000;
constexpr uint32_t pongTimeout = 5000;
uint8_t disconnectTimeoutCount = 3;

ServerConnection::ServerConnection(const char* url, int port, const char* path) : 
    url(url), port(port), path(path) {

}

void ServerConnection::connect(const char* key, uint32_t reconnectTime) {
    webSocket.begin(url, port, path);
    registerEventCallback();
    reconnectAfter(reconnectTime);
}

void ServerConnection::connectSSL(const char* key, uint32_t reconnectTime) {
    if (socketIsConnected) {
        return;
    }
    this->key = key;
    webSocket.beginSSL(url, port, path);
    registerEventCallback();
    reconnectAfter(reconnectTime);
}

void ServerConnection::loop() {
    webSocket.loop();
}

void ServerConnection::onMessage(MessageCallback callback) {
    messageCallback = callback;
}

void ServerConnection::reconnectAfter(uint32_t reconnectTime) {
    webSocket.setReconnectInterval(reconnectTime);
}

void ServerConnection::registerEventCallback() {
    std::function<void(WStype_t, uint8_t *, size_t)> f = [this](WStype_t type, uint8_t *payload, size_t length) {
        this->webSocketEventHandler(type, payload, length);
    };
    webSocket.onEvent(f);
}

void ServerConnection::webSocketEventHandler(WStype_t type, uint8_t * payload, size_t length) {
switch(type) {
    case WStype_DISCONNECTED:
        socketIsConnected = false;
        Serial.println("[INFO] Socket disconnected.");
        break;
    case WStype_CONNECTED:
        socketIsConnected = true;
        webSocket.sendTXT(key);
        Serial.printf("[INFO] Socket connected to url: %s\n", payload);
        webSocket.enableHeartbeat(pingInterval, pongTimeout, disconnectTimeoutCount);
        break;
    case WStype_TEXT:
        sendFailureResponse(SesameEvent::TextReceived);
        break;
    case WStype_BIN:
        processReceivedBytes(payload, length);
        break;
    case WStype_PONG:
        break;
    case WStype_PING:
    case WStype_ERROR:
    case WStype_FRAGMENT_TEXT_START:
    case WStype_FRAGMENT_BIN_START:
    case WStype_FRAGMENT:
    case WStype_FRAGMENT_FIN:
        sendFailureResponse(SesameEvent::UnexpectedSocketEvent);
        break;
    }
}

void ServerConnection::processReceivedBytes(uint8_t* payload, size_t length) {
    if (length != AUTHENTICATED_MESSAGE_SIZE) {
        sendFailureResponse(SesameEvent::InvalidMessageData);
        return;
    }
    AuthenticatedMessage* message = (AuthenticatedMessage*) payload;
    if (messageCallback == NULL) {
        sendFailureResponse(SesameEvent::MessageAuthenticationFailed);
        return;
    }
    AuthenticatedMessage responseMessage;
    SesameEvent event = messageCallback(message, &responseMessage);
    sendResponse(event, &responseMessage);
}

void ServerConnection::sendFailureResponse(SesameEvent event) {
    uint8_t response = static_cast<uint8_t>(event);
    webSocket.sendBIN(&response, 1);
    Serial.printf("[INFO] Socket failure %d\n", response);
}

void ServerConnection::sendResponse(SesameEvent event, AuthenticatedMessage* message) {
    uint8_t response[AUTHENTICATED_MESSAGE_SIZE+1];
    response[0] = static_cast<uint8_t>(event);
    memcpy(response+1, (uint8_t*) message, AUTHENTICATED_MESSAGE_SIZE);
    webSocket.sendBIN(response, AUTHENTICATED_MESSAGE_SIZE+1);
    Serial.printf("[INFO] Socket response %d\n", response[0]);
}